﻿using System;
using System.Data.Entity;
using GameStore.Entities.DataAccessModels;

namespace GameStore.Entities.Contexts
{
    /// <summary>
    /// Describes part of database dedicated to the store
    /// </summary>
    internal class GameStoreContext : DbContext
    {
                /// <summary>
        /// Exposes logging capabilities to <see cref="Repositories.GameStoreRepository"/>.
        /// </summary>
        internal Action<string> Log { get; set; }


        /// <summary>
        /// This constructor is required by migrations
        /// </summary>
        public GameStoreContext()
            : base()
        {
        }


        /// <summary>
        /// This constructors should be used in code
        /// </summary>
        /// <param name="nameOrConnectionString"></param>
        public GameStoreContext(string nameOrConnectionString)
            : base(nameOrConnectionString)
        {
            // safely invokes Log property action only if specified
            Database.Log = log => Log?.Invoke(log);
        }


        /// <summary>
        /// Method specifies specific database details related to the used models
        /// </summary>
        protected override void OnModelCreating(DbModelBuilder modelBuilder)
        {
            // Rename table named by type
            modelBuilder
                .Entity<GameWrapper>()
                .ToTable("Games");

            // Solve M:N relation
            modelBuilder
                .Entity<Award>()
                .HasMany(award => award.Games)
                .WithMany(game => game.Awards)                
                .Map(map => map.ToTable("GameAwards"));

            // Solve 1:N relation
            modelBuilder
                .Entity<Developer>()
                .HasMany(developer => developer.Games)
                .WithRequired(game => game.Developer)
                .WillCascadeOnDelete();

            // Base implementation is empty, however, this is not necessarily granted
            base.OnModelCreating(modelBuilder);
        }

        // Each relevant non-relational database table should be specified by a DbSet

        public virtual DbSet<Award> Awards { get; set; }

        public virtual DbSet<Developer> Developers { get; set; }

        public virtual DbSet<GameWrapper> Games { get; set; }
    }
}
