﻿using System;
using System.Collections.Generic;
using System.Data.Entity;
using System.Linq;
using System.Threading.Tasks;
using AutoMapper;
using GameStore.Entities.Configuration;
using GameStore.Entities.Contexts;
using GameStore.Entities.DataAccessModels;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;

namespace GameStore.Entities.Repositories
{
    /// <summary>
    /// Exposes <see cref="Game"/> and its related entities (<see cref="Award"/> and <see cref="Developer"/>) without exposing entity framework.
    /// </summary>
    public class GameStoreRepository : IGameStoreRepository
    {
        private GameStoreContext _storeContext;
        private IMapper _mapper;

        /// <summary>
        /// Constructor indended for tests' purposes only.
        /// </summary>
        /// <param name="gameStoreContext">Instance of (eventually mocked) DbContext</param>
        /// <param name="mapper">Automapper with cofigured mappings</param>
        internal GameStoreRepository(GameStoreContext gameStoreContext, IMapper mapper)
        {
            if (gameStoreContext == null)
                throw new ArgumentNullException(nameof(gameStoreContext));
            if (mapper == null)
                throw new ArgumentNullException(nameof(mapper));

            _storeContext = gameStoreContext;
            _mapper = mapper;
        }

        /// <summary>
        /// Default repository constructor that uses connection string provided by top most level configuration
        /// </summary>
        public GameStoreRepository(IOptions<ConnectionOptions> options, ILogger<IGameStoreRepository> logger, IMapper mapper)
            : this(new GameStoreContext(options.Value.ConnectionString), mapper)
        {
            // To show these messages in console, default log level needs to be set to Debug
            _storeContext.Log = message => logger.LogDebug(message);
        }

        public async Task<GameWrapper> AddGame(GameWrapper game)
        {
            if (game == null)
                throw new ArgumentNullException(nameof(game));

            var addedGame = _storeContext.Games.Add(game);
            await _storeContext.SaveChangesAsync();
            return addedGame;
        }

        public async Task<GameWrapper> AddGame(Game game, Developer developer, IEnumerable<Award> awards)
        {
            if (game == null)
                throw new ArgumentNullException(nameof(game));
            if (developer == null)
                throw new ArgumentNullException(nameof(developer));
            if (awards == null)
                throw new ArgumentNullException(nameof(awards));

            GameWrapper wrappedGame = _mapper.Map<Game, GameWrapper>(game);

            // To prevent re-edition of existing object, existing objects is first cached from DB
            wrappedGame.Developer = _storeContext.Developers.Find(developer.Id);

            awards
                .Select(award => _storeContext.Awards.Find(award.Id))
                .ToList()
                .ForEach(wrappedGame.Awards.Add);

            // Stores new game, expecting both provider and awards to exist
            var addedGame = _storeContext.Games.Add(wrappedGame);
            await _storeContext.SaveChangesAsync();

            return addedGame;
        }

        public async Task<GameWrapper> GetSingleGame(string codeName)
            => await _storeContext
                .Games
                .Include(wrapper => wrapper.Developer)
                .Include(wrapper => wrapper.Awards)
                .Where(game => game.CodeName == codeName)
                .SingleOrDefaultAsync();

        public async Task<IEnumerable<Game>> GetAllGames()
            => _mapper.Map<IEnumerable<GameWrapper>, IEnumerable<Game>>(
                await _storeContext.Games.ToArrayAsync());
    }
}
