﻿using System;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.Infrastructure;
using Microsoft.AspNetCore.Mvc.Routing;
using Microsoft.Extensions.Logging;
using GameStore.Services.DataTransferModels;
using GameStore.Services.GameStock;
using GameStore.Services.RandomId;

namespace GameStore.Controllers.Api
{
    [Route("api/[controller]")]
    //[Route("api/games")]
    public class GamesController : Controller
    {
        private readonly IGameStockService _gameStockService;
        private readonly IRandomIdService _randomId;
        private readonly IActionContextAccessor _actionContextAccessor;
        private readonly IHttpContextAccessor _httpContextAccessor;
        private readonly IUrlHelperFactory _urlHelperFactory;
        private readonly ILogger<GamesController> _logger;

        public GamesController(IGameStockService gameStockService, IRandomIdService randomId, IActionContextAccessor actionContextAccessor, IHttpContextAccessor httpContextAccessor, IUrlHelperFactory urlHelperFactory, ILogger<GamesController> logger)
        {
            _gameStockService = gameStockService;
            _randomId = randomId;
            _actionContextAccessor = actionContextAccessor;
            _httpContextAccessor = httpContextAccessor;
            _urlHelperFactory = urlHelperFactory;
            _logger = logger;
        }

        // GET api/games
        [HttpGet]
        //public async Task<IActionResult> GetAllGames(
        public async Task<IActionResult> GetAllGames(
            [FromQuery(Name = "limit")]int limit,
            [FromQuery(Name = "attributes")]string[] attributes)
        {
            var games = await _gameStockService.GetAllGames();

            // Apply limit
            if (limit > 0)
            {
                games = games.Take(limit).ToArray();
            }

            // Select only given fields
            if (attributes != null && attributes.Any())
            {
                games = games.Select(game =>
                {
                    var projection = new GameItem();

                    foreach (var attribute in attributes)
                    {
                        projection.GetType()
                            .GetProperty(attribute)
                            .SetValue(projection, game.GetType().GetProperty(attribute).GetValue(game));
                    }

                    return projection;
                }).ToArray();
            }

            _logger.LogInformation($"Retrieved {games.Length} games.");
            return Ok(games);
        }

        // GET api/games/wow
        [HttpGet("{codename}")]
        public async Task<IActionResult> GetGame(string codename)
        {
            var game = await _gameStockService.GetGame(codename);
            return Ok(game);
        }

        // POST api/games
        [HttpPost("{developerCodeName}")]
        public async Task<IActionResult> Post([FromBody]GameItem game, [FromRoute]string developerCodeName)
        {
            try
            {
                var result = await _gameStockService.AddGame(game, developerCodeName);

                var urlHelper = _urlHelperFactory.GetUrlHelper(_actionContextAccessor.ActionContext);
                var newUrl = new Uri(urlHelper.Action("GetGame", "Games", new
                {
                    codename = game.GameCodeName,
                }, _httpContextAccessor.HttpContext.Request.Scheme));
                _logger.LogInformation("Generated new game with name " + game.DisplayName);

                return Created(newUrl, result);
            }
            catch (Exception ex)
            {
                return BadRequest(ex.Message);
            }
        }

        // PUT api/games/wow
        [HttpPut("{id}")]
        public void Put(int id, [FromBody]string value)
        {
        }

        // DELETE api/games/wow
        [HttpDelete("{id}")]
        public void Delete(int id)
        {
        }
    }
}
