﻿using System;
using System.Collections.Generic;
using System.Data.Entity;
using System.Data.Entity.Infrastructure;
using System.Linq;
using AutoMapper;
using GameStore.Entities.Configuration;
using GameStore.Entities.Contexts;
using GameStore.Entities.DataAccessModels;
using GameStore.Entities.Repositories;
using NSubstitute;
using NSubstitute.Core;
using Xunit;

namespace GameStore.Entities.UnitTests
{
    public class GameStoreRepositoryTests
    {
        private GameStoreRepository _GameStoreRepository;

        public static IEnumerable<object[]> InvalidAddGameArguments
        {
            get
            {
                yield return new object[] { (Game)null, new Developer(), Enumerable.Empty<Award>() };
                yield return new object[] { new Game(), (Developer)null, Enumerable.Empty<Award>() };
                yield return new object[] { new Game(), new Developer(), (IEnumerable<Award>)null };
            }
        }

        [Fact]
        public async void GetSingleGame_Existing_ReturnsGame()
        {
            var actualGame = await _GameStoreRepository.GetSingleGame("TestGame1");

            Assert.NotNull(actualGame);
            Assert.IsType<Game>(actualGame);
        }

        [Theory]
        [InlineData(null)]
        [InlineData("")]
        public async void GetSingleGame_NullOrEmpty_ReturnsGame(string codeName)
        {
            var actualGame = await _GameStoreRepository.GetSingleGame(codeName);

            Assert.Null(actualGame);
        }

        [Fact]
        public async void GetAllGames_ReturnsAllGameInstances()
        {
            var actualGames = await _GameStoreRepository.GetAllGames();

            Assert.NotNull(actualGames);
            Assert.NotEmpty(actualGames);
            Assert.Equal(4, actualGames.Count());
            foreach (var actualGame in actualGames)
            {
                Assert.IsType<Game>(actualGame);
            }
        }

        [Fact]
        public async void AddGame_NoWrapper_Throws()
        {
            await Assert.ThrowsAsync<ArgumentNullException>(() => _GameStoreRepository.AddGame(null));
        }

        [Fact]
        public async void AddGame_CorrectWrapper_ReturnsGameWithId()
        {
            var newGame = new GameWrapper
            {
                CodeName = "NewTestGame",
                DisplayName = "New Test Game",
                Price = 123.4m,
                Developer = GetTestDeveloperWithoutGames(128),
                Awards = {
                    GetTestAwardWithoutGames(128),
                    GetTestAwardWithoutGames(129),
                    GetTestAwardWithoutGames(130),
                    GetTestAwardWithoutGames(131),
                    GetTestAwardWithoutGames(132)
                }
            };
            Game actualGame = await _GameStoreRepository.AddGame(newGame);

            // Checks returned value
            Assert.NotNull(actualGame);
            Assert.Equal(newGame.CodeName, actualGame.CodeName);
            Assert.Equal(newGame.DisplayName, actualGame.DisplayName);
            Assert.NotEqual(Guid.Empty, actualGame.Id);

            // Check storage is persistant
            Assert.NotNull(_GameStoreRepository.GetSingleGame(newGame.CodeName));
        }

        [Theory]
        [MemberData(nameof(InvalidAddGameArguments))]
        public async void AddGame_NullGameOrDeveloperOrAwards_Throws(Game game, Developer developer, IEnumerable<Award> awards)
        {
            await Assert.ThrowsAsync<ArgumentNullException>(() => _GameStoreRepository.AddGame(game, developer, awards));
        }

        [Fact]
        public async void AddGame_CorrectParts_ReturnsGameWithId()
        {
            var newGame = new Game
            {
                CodeName = "AnotherNewTestGame",
                DisplayName = "Another New Test Game",
                Price = 432.1m,
            };
            var developer = GetTestDeveloperWithoutGames(255);
            var awards = new[]
            {
                    GetTestAwardWithoutGames(254),
                    GetTestAwardWithoutGames(253),
                    GetTestAwardWithoutGames(252),
            };

            Game actualGame = await _GameStoreRepository.AddGame(newGame, developer, awards);

            // Checks returned value
            Assert.NotNull(actualGame);
            Assert.Equal(newGame.CodeName, actualGame.CodeName);
            Assert.Equal(newGame.DisplayName, actualGame.DisplayName);
            Assert.NotEqual(Guid.Empty, actualGame.Id);

            // Checks storage is persistant
            Assert.NotNull(_GameStoreRepository.GetSingleGame(newGame.CodeName));
        }

        /// <summary>
        /// https://msdn.microsoft.com/en-us/library/dn314429.aspx
        /// List has to be used in order for addition to be possible.
        /// </summary>
        public GameStoreRepositoryTests()
        {
            // Mock awards
            Award testAward1 = GetTestAwardWithoutGames(1);
            Award testAward2 = GetTestAwardWithoutGames(2);

            var awardsData = new List<Award>
            {
                testAward1,
                testAward2
            };

            // Mock game developers
            Developer testDeveloper1 = GetTestDeveloperWithoutGames(1);
            Developer testDeveloper2 = GetTestDeveloperWithoutGames(2);

            var developersData = new List<Developer>{
                testDeveloper1,
                testDeveloper2
            };

            // Mock games
            var gamesData = new List<GameWrapper>
            {
                GetTestGameWrapper(1, testDeveloper1, testAward1, testAward2),
                GetTestGameWrapper(2, testDeveloper2),
                GetTestGameWrapper(3, testDeveloper2, testAward2),
                GetTestGameWrapper(4, testDeveloper1, testAward1),
            };

            // Mock context
            var awardsSubstitute = SubstituteQueryable(awardsData);
            var developersSubstitute = SubstituteQueryable(developersData);
            var gamesSubstitute = SubstituteQueryable(gamesData);

            var contextSubsitute = Substitute.For<GameStoreContext>();
            contextSubsitute.Awards.Returns(awardsSubstitute);
            contextSubsitute.Developers.Returns(developersSubstitute);
            contextSubsitute.Games.Returns(gamesSubstitute);

            // Mock repository, use real mapper
            var mapper = new MapperConfiguration(EntitiesMapperConfiguration.InitialializeMappings).CreateMapper();
            _GameStoreRepository = new GameStoreRepository(contextSubsitute, mapper);
        }

        /// <summary>
        /// Creates substitute for a <see cref="DbSet{TEntity}"/> with database replaced with an in-memory structure represented by <paramref name="data"/>.
        /// Can be used for querying and addition, including async operations.
        /// </summary>
        /// <typeparam name="TType">Type of data and <see cref="DbSet{TEntity}"/> to substitute</typeparam>
        /// <param name="data">Initial content of "database"</param>
        /// <returns>Queryable that can be used as <see cref="DbSet{TEntity}"/> substitute.</returns>
        private static IQueryable<TType> SubstituteQueryable<TType>(ICollection<TType> data)
            where TType : ModelBase
        {
            var queryableData = data.AsQueryable();
            var queryableSubstitute = Substitute.For<IQueryable<TType>, IDbAsyncEnumerable<TType>, DbSet<TType>>();

            // Mock queryable
            queryableSubstitute.Provider.Returns(new TestDbAsyncQueryProvider<TType>(queryableData.Provider));
            queryableSubstitute.Expression.Returns(queryableData.Expression);
            queryableSubstitute.ElementType.Returns(queryableData.ElementType);
            queryableSubstitute.GetEnumerator().Returns(queryableData.GetEnumerator());

            // Mock addition
            ((DbSet<TType>)queryableSubstitute).Add(null).ReturnsForAnyArgs(callInfo => SimulateAddition(callInfo, data));

            // Mock async
            ((IDbAsyncEnumerable<TType>)queryableSubstitute).GetAsyncEnumerator().Returns(new TestDbAsyncEnumerator<TType>(data.GetEnumerator()));

            return queryableSubstitute;
        }

        /// <summary>
        /// Reads <typeparamref name="TType"/> from <paramref name="callInfo"/> and stores it to the <paramref name="data"/>.
        /// To emulate reald DB, it also sets <see cref="ModelBase.Id"/> with new <see cref="Guid"/> and returns the very
        /// object the method was provided with.
        /// </summary>
        private static TType SimulateAddition<TType>(CallInfo callInfo, ICollection<TType> data)
            where TType : ModelBase
        {
            TType entry = callInfo.Arg<TType>();

            entry.Id = Guid.NewGuid();
            data.Add(callInfo.Arg<TType>());

            return entry;
        }

        /// <summary>
        /// Creates new test <see cref="GameWrapper"/> object with <paramref name="developer"/> and <paramref name="awards"/>.
        /// Object name and identifier is accompanied with <paramref name="identifier"/>.
        /// </summary>
        private static GameWrapper GetTestGameWrapper(int identifier, Developer developer, params Award[] awards)
        {
            var gameWrapper = new GameWrapper
            {
                CodeName = $"TestGame{identifier}",
                DisplayName = $"Test Game {identifier}",
                Price = 1,
                Id = new Guid($"{{00000000-0000-0000-1111-{FormatForLastGuidPart(identifier)}}}"),
                Developer = developer
            };

            developer.Games.Add(gameWrapper);
            foreach(var award in awards)
            {
                gameWrapper.Awards.Add(award);
                award.Games.Add(gameWrapper);
            }

            return gameWrapper;
        }

        /// <summary>
        /// Creates new test <see cref="Developer"/> object with no <see cref="Developer.Games"/>.
        /// Object name and identifier is accompanied with <paramref name="identifier"/>.
        /// </summary>
        private static Developer GetTestDeveloperWithoutGames(int identifier)
        {
            return new Developer
            {
                CodeName = $"TestDeveloper{identifier}",
                DisplayName = $"Test Developer {identifier}",
                Id = new Guid($"{{00000000-0000-1111-0000-{FormatForLastGuidPart(identifier)}}}")
            };
        }

        /// <summary>
        /// Creates new test <see cref="Award"/> object with no <see cref="Award.Games"/>.
        /// Object name and identifier is accompanied with <paramref name="identifier"/>.
        /// </summary>
        private static Award GetTestAwardWithoutGames(int identifier)
        {
            return new Award
            {
                CodeName = $"TestAward{identifier}",
                DisplayName = $"Test Award {identifier}",
                Id = new Guid($"{{00000000-1111-0000-0000-{FormatForLastGuidPart(identifier)}}}")
            };
        }

        /// <summary>
        /// Formates given <paramref name="identifier"/> to (at least) 12 characters long string where all missing characters are replaced with 0.
        /// <see cref="String"/> formatted in this way can be used as last part a GUID.
        /// </summary>
        private static string FormatForLastGuidPart(int identifier) => identifier.ToString().PadLeft(12, '0');
    }
}
